#include "SymbolTable.h"
#include <iomanip>
#include <iostream>
#include <sstream>
#include "Type.h"

bool SymbolEntry::setNext(SymbolEntry* se) {
    SymbolEntry* s = this;
    long unsigned int cnt =
        ((FunctionType*)(se->getType()))->getParamsType().size();
    if (cnt == ((FunctionType*)(s->getType()))->getParamsType().size())
        return false;
    while (s->getNext()) {
        if (cnt == ((FunctionType*)(s->getType()))->getParamsType().size())
            return false;
        s = s->getNext();
    }
    if (s == this) {
        this->next = se;
    } else {
        s->setNext(se);
    }
    return true;
}

SymbolEntry::SymbolEntry(Type* type, int kind) {
    this->type = type;
    this->kind = kind;
}

ConstantSymbolEntry::ConstantSymbolEntry(Type* type, double value)
    : SymbolEntry(type, SymbolEntry::CONSTANT) {
    assert(type->isInt() || type->isFloat());
    this->value = value;
    //std::cout<<"new constantsymbolentry1\n";
}

ConstantSymbolEntry::ConstantSymbolEntry(Type* type)
    : SymbolEntry(type, SymbolEntry::CONSTANT) {
    // assert(type->isArray());
    //std::cout<<"new constantsymbolentry3\n";
}

double ConstantSymbolEntry::getValue() const {
    assert(type->isInt() || type->isFloat());
    return value;
}

std::string ConstantSymbolEntry::toStr() {
    std::ostringstream buffer;
    if (type->isInt()) {
        if (value == 2147483648)
            buffer << "2147483648";
        else
            buffer << (int)value;
    } else if (type->isFloat())
        buffer << std::fixed << value;
    return buffer.str();
}

IdentifierSymbolEntry::IdentifierSymbolEntry(Type* type,
                                             std::string name,
                                             int scope,
                                             int paramNo,
                                             bool sysy)
    : SymbolEntry(type, SymbolEntry::VARIABLE),
      name(name),
      sysy(sysy),
      paramNo(paramNo) {
    this->scope = scope;
    this->initial = false;
    this->label = -1;
    this->allZero = false;
    this->constant = false;
    this->notZeroNum = 0;
    this->param2local = false;
}

void IdentifierSymbolEntry::setValue(double value) {
    if (((IntType*)(this->getType()))->isConst()) {  // IntType is ok here.
        if (!initial) {
            this->value = value;
            initial = true;
        } else {
            fprintf(stderr, "trying to set value for constant.\n");
        }
    } else {
        this->value = value;
    }
}

void IdentifierSymbolEntry::setArrayValue(double* arrayValue) {
    if (((ArrayType*)(this->getType()))->isConst()) {
        if (!initial) {
            this->arrayValue = arrayValue;
            initial = true;
        } else {
            fprintf(stderr, "trying to set value for constant.\n");
        }
    } else {
        this->arrayValue = arrayValue;
    }
}

std::string IdentifierSymbolEntry::toStr() {
    std::ostringstream buffer;
    if (label < 0) {
        if (type->isFunc())
            buffer << '@';
        buffer << name;
    } else
        buffer << "%t" << label;
    return buffer.str();
}

TemporarySymbolEntry::TemporarySymbolEntry(Type* type, int label)
    : SymbolEntry(type, SymbolEntry::TEMPORARY) {
    this->label = label;
}

std::string TemporarySymbolEntry::toStr() {
    std::ostringstream buffer;
    buffer << "%t" << label;
    return buffer.str();
}

SymbolTable::SymbolTable() {
    prev = nullptr;
    level = 0;
}

SymbolTable::SymbolTable(SymbolTable* prev) {
    this->prev = prev;
    this->level = prev->level + 1;
}

/*
    Description: lookup the symbol entry of an identifier in the symbol table
    Parameters:
        name: identifier name
    Return: pointer to the symbol entry of the identifier

    hint:
    1. The symbol table is a stack. The top of the stack contains symbol entries
   in the current scope.
    2. Search the entry in the current symbol table at first.
    3. If it's not in the current table, search it in previous ones(along the
   'prev' link).
    4. If you find the entry, return it.
    5. If you can't find it in all symbol tables, return nullptr.
*/
SymbolEntry* SymbolTable::lookup(std::string name) {
    SymbolTable* table = this;
    while (table != nullptr)
        if (table->symbolTable.find(name) != table->symbolTable.end()) {
            return table->symbolTable[name];
        } else {
            table = table->prev;
        }
    return nullptr;
}

// install the entry into current symbol table.
bool SymbolTable::install(std::string name, SymbolEntry* entry) {
    // 同时检查是否重定义
    if (this->symbolTable.find(name) != this->symbolTable.end()) {
        SymbolEntry* se = this->symbolTable[name];
        if (se->getType()->isFunc())
            return se->setNext(entry);
        return false;
    } else {
        symbolTable[name] = entry;
        return true;
    }
}

bool SymbolTable::installSysFunc(std::string name) {
    std::vector<Type*> typeVec;
    std::vector<SymbolEntry*> paramVec;
    FunctionType* funcType;
    if(name == "getint") {
        funcType = new FunctionType(TypeSystem::intType, typeVec, paramVec);
    } else if(name == "getch") {
        funcType = new FunctionType(TypeSystem::intType, typeVec, paramVec);
    } else if(name == "getfloat") {
        funcType = new FunctionType(TypeSystem::floatType, typeVec, paramVec);
    } else if(name == "_sysy_starttime") {
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "_sysy_stoptime"){
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "getarray") {
        ArrayType* arr = new ArrayType(TypeSystem::intType, -1);
        typeVec.push_back(arr);
        funcType = new FunctionType(TypeSystem::intType, typeVec, paramVec);
    } else if(name == "getfarray") {
        ArrayType* arr = new ArrayType(TypeSystem::floatType, -1);
        typeVec.push_back(arr);
        funcType = new FunctionType(TypeSystem::intType, typeVec, paramVec); 
    } else if(name == "putint") {
        typeVec.push_back(TypeSystem::intType);
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "putch") {
        typeVec.push_back(TypeSystem::intType);
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "putfloat"){
        typeVec.push_back(TypeSystem::floatType);
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "putarray") {
        typeVec.push_back(TypeSystem::intType);
        ArrayType* arr = new ArrayType(TypeSystem::intType, -1);
        typeVec.push_back(arr);
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } else if(name == "putfarray") {
        typeVec.push_back(TypeSystem::intType);
        ArrayType* arr = new ArrayType(TypeSystem::floatType, -1);
        typeVec.push_back(arr);
        funcType = new FunctionType(TypeSystem::voidType, typeVec, paramVec);
    } 
    IdentifierSymbolEntry* id = new IdentifierSymbolEntry(funcType, name, 0, -1, true);
    int intParamNum = 0, floatParamNum = 0;
    for(auto it = funcType->getParamsType().begin(); it != funcType->getParamsType().end(); it++){
        if(*it == TypeSystem::intType){
            intParamNum++;
        }else if(*it == TypeSystem::floatType){
            floatParamNum++;
        }
    }
    id->setIntParamNo(intParamNum);
    id->setFloatParamNo(floatParamNum);
    return this->install(name, id);
}

int SymbolTable::counter = 0;
static SymbolTable t;
SymbolTable* identifiers = &t;
SymbolTable* globals = &t;
